class JobStep < ActiveRecord::Base

  include Common::EarPassword
  include Common::ProxyPassword
  include Common::Obfuscator
  include JobAndStepCommon::Category

  concerned_with :action_request, :background, :email, :map_and_chart, :queueing, :security, :soap, :transfer_request

  belongs_to :job
  belongs_to :source_login, :class_name => "Login", :foreign_key => 'source_saved_login_id'
  belongs_to :dest_login, :class_name => "Login", :foreign_key => 'dest_saved_login_id'
  belongs_to :source_node,    :class_name => "FaspNode",    :foreign_key => "source_node_id"
  belongs_to :dest_node,      :class_name => "FaspNode",    :foreign_key => "dest_node_id"
  belongs_to :fasp_session
  belongs_to :client_session, :class_name => "FaspSession", :foreign_key => "client_fasp_session_id"
  belongs_to :server_session, :class_name => "FaspSession", :foreign_key => "server_fasp_session_id"
  has_many :job_step_users # view
  has_many :sessions, :order => "started_at", :class_name => "FaspSession", :foreign_key => :job_step_id, :dependent => :nullify

  has_many :client_sessions, :order => "started_at", :class_name => "FaspSession", :foreign_key => :job_step_id, :conditions => {:initiator => FaspSession::INITIATOR_LOCAL}
  has_many :server_sessions, :order => "started_at", :class_name => "FaspSession", :foreign_key => :job_step_id, :conditions => {:initiator => FaspSession::INITIATOR_REMOTE}

  has_many :emails, :class_name => "JobStepEmail", :dependent => :destroy do
    # Returns Array of JobStepEmail's matching _type_.
    def for_type(type)
      self.select do |jse|
        method = "#{type}_template_id?"
        jse.send(method)
      end
    end
  end
  
  has_many :email_events, :class_name => 'JobStepEmailEvent', :dependent => :destroy
  
  has_many :sent_email_events do
    # Returns +true+ if a "start" email has been sent for this step.
    def start?
      email_types.include? JobEmail::TYPE_START
    end
    
    # Returns +true+ if a "success" email has been sent for this step.
    def success?
      email_types.include? JobEmail::TYPE_SUCCESS
    end
    
    # Returns +true+ if an "error" email has been sent for this step.
    def error?
      email_types.include? JobEmail::TYPE_ERROR
    end

    # Returns (cached) array of sent email types.
    def email_types
      @email_types ||= map(&:email_type)
    end
  end

  validate :validate_archive_directory

  def validate_archive_directory
    if self.archive_after_transfer
      errors.add_to_base("Archive directory must be specified if archive after transfer is selected") unless self.archive_directory.present?
    end
  end

  before_create :select_best_initial_endpoints, :if => :should_select_initial_endpoints?

  after_create :create_email_events
  before_update :update_email_events

  before_save :encrypt_ear_password
  before_save :encrypt_proxy_password

  STATUS_QUEUED        = Console::STATUS_QUEUED
  STATUS_SUBMITTED     = Console::STATUS_SUBMITTED
  STATUS_SCHEDULED     = Console::STATUS_SCHEDULED
  #~ STATUS_STARTED       = DbLogger::STATUS_STARTED  -- JobStep should never be assigned status STARTED, only RUNNING
  STATUS_RUNNING       = DbLogger::STATUS_RUNNING
  STATUS_PAUSED        = DbLogger::STATUS_PAUSED
  STATUS_UNKNOWN       = Console::STATUS_UNKNOWN
  STATUS_COMPLETED     = DbLogger::STATUS_COMPLETED
  STATUS_ERROR         = DbLogger::STATUS_ERROR
  STATUS_WILLRETRY     = DbLogger::STATUS_WILLRETRY
  STATUS_CANCELLED     = DbLogger::STATUS_CANCELLED
  STATUS_INACTIVE      = Console::STATUS_INACTIVE
  STATUS_VALIDATING    = DbLogger::STATUS_VALIDATING
  STATUS_TO_BE_VALIDATED = DbLogger::STATUS_TO_BE_VALIDATED
#  STATUS_NEVER_STARTED = Console::STATUS_NEVER_STARTED
  
  STATUSES_ALL = [STATUS_QUEUED, STATUS_SUBMITTED, STATUS_SCHEDULED, STATUS_RUNNING, STATUS_PAUSED,
                  STATUS_UNKNOWN,   STATUS_COMPLETED, STATUS_ERROR, STATUS_WILLRETRY, STATUS_CANCELLED,
                  STATUS_VALIDATING, STATUS_TO_BE_VALIDATED, STATUS_INACTIVE] #,  STATUS_NEVER_STARTED]
                  
  # used to determine which steps should be updated from session rows
  STATUSES_POSSIBLY_ACTIVE = Console::STATUSES_POSSIBLY_ACTIVE_FOR_JOB_AND_STEP
  
  STATUSES_ACTIVE = [Console::STATUS_STARTED, STATUS_RUNNING, STATUS_PAUSED]
  
  # statuses for which 'success' email notifications will be sent
  STATUSES_SUCCESS = [STATUS_COMPLETED]
  
  # statuses for which 'error' email notifications will be sent
  # (for status 'will_retry', a certain timeout period must pass before error email is sent)
  STATUSES_ERROR = [STATUS_ERROR, STATUS_WILLRETRY, STATUS_CANCELLED, STATUS_INACTIVE]
  
  # statuses_started = statuses_all - statuses_ended
  # FIXME: NEEDED?
  STATUSES_ENDED = [
    # cancelled
    # no longer active
    STATUS_COMPLETED,
    STATUS_ERROR
  ]

  RETRY_TIMEOUT_ERR_MSG = "Timeout waiting for retry."

#  EXIT_CODE_SCHEDULED_NEVER_STARTED = 322
#  ERR_DESC_SCHEDULED_NEVER_STARTED = 'scheduled step never started'

  named_scope :ui,        :conditions => {:marked_for_delete => false}
  named_scope :willretry, :conditions => {:status => STATUS_WILLRETRY}
  # guard against session getting deleted by background process with session ? x : y, where y will always evaluate to false
  
  named_scope :last_day, lambda {{:conditions => ["started_at > ? or ended_at > ? or ended_at is null", 1.day.ago, 1.day.ago]}}
  named_scope :active, :conditions => {:status => STATUSES_ACTIVE}
  named_scope :possibly_active, :conditions => {:status => STATUSES_POSSIBLY_ACTIVE}
  named_scope :scheduled, :conditions => {:status => STATUS_SCHEDULED}
  
  named_scope :to_recalc, :conditions => ["force_recalc = ? OR status IN (?)", true, Console::STATUSES_POSSIBLY_ACTIVE_FOR_JOB_AND_STEP]
  named_scope :marked_and_ready_for_delete, lambda{{:conditions => ["marked_for_delete = 1 and ended_at < ?", Console::TRANSFER_DATA_DELETION_WAIT.ago]}}

  named_scope :failover_possible_scope, :conditions => ["failover_possible = true"]

  # reimplement named_scope in pure sql for performance
  def self.to_recalc_sql
    sql = "SELECT * FROM job_steps WHERE (force_recalc = 1) UNION DISTINCT SELECT * FROM job_steps WHERE status IN ('queued','submitted','running','paused','unknown','validating','to_be_validated') ;"
    find_by_sql(sql)
  end


  def initialize(attributes=nil)
    super
    self.uuid = UUID.random_create.to_s
  end

  def validating?
    [STATUS_TO_BE_VALIDATED, STATUS_VALIDATING].include?(self.status)
  end

  def submittable_via_node_api?
    source_login.present? && source_login.node_api_based_login? &&
      dest_login.present? && dest_login.node_api_based_login?
  end

  # Submittable via one-sided node API request
  # One side is a managed node with node API login and the other is an SSH login
  def submittable_via_direct_request?
    return unless source_login.present? && dest_login.present?

    (source_login.node_api_based_login? && source_login.node.present? && source_login.node.is_a?(FaspNode) && dest_login.ssh_login?)
  end

  def shares_transfer?
    source_login.present? && source_login.shares_login? ||
      dest_login.present? && dest_login.shares_login?
  end

  def source_session
    if client_session && client_session.outbound?
      client_session
    elsif server_session && server_session.outbound?
      server_session
    else
      nil
    end
  end

  def dest_session
    if client_session && client_session.inbound?
      client_session
    elsif server_session && server_session.inbound?
      server_session
    else
      nil
    end
  end

  def access_key_display_name(client_or_server)
    return unless client_or_server == 'client' || client_or_server == 'server'

    name = self.send("#{client_or_server}_access_key_name")
    key = self.send("#{client_or_server}_access_key")

    name.present? ? "#{name} (#{key})" : key
  end

  def source_access_key
    source_initiated? ? client_access_key : server_access_key
  end

  def source_access_key_display_name
    source_initiated? ? access_key_display_name('client') : access_key_display_name('server')
  end

  def dest_access_key
    source_initiated? ? server_access_key : client_access_key
  end

  def dest_access_key_display_name
    source_initiated? ? access_key_display_name('server') : access_key_display_name('client')
  end

  def source_docroot
    if source_session
      if source_session.local?
        source_session.client_docroot
      else
        source_session.server_docroot
      end
    else
      nil
    end
  end
  
  def dest_docroot
    if dest_session
      if dest_session.local?
        dest_session.client_docroot
      else
        dest_session.server_docroot
      end
    else
      nil
    end
  end

  def proxy_url_with_credentials
    if use_fasp_proxy?
      if proxy_user.present?
        if proxy_unencrypted_password.present?
          "#{proxy_protocol}://#{proxy_user}:#{proxy_unencrypted_password}@#{proxy_host}:#{proxy_port}"
        else
          "#{proxy_protocol}://#{proxy_user}@#{proxy_host}:#{proxy_port}"
        end
      else
        proxy_url
      end
    end
  end

  def proxy_url
    if use_fasp_proxy?
      "#{proxy_protocol}://#{proxy_host}:#{proxy_port}"
    end
  end

  def proxy_protocol
    proxy_secure? ? 'dnats' : 'dnat'
  end

  # return list of session records but
  # if we have data from both sides of the transfer,
  # give preference to source side data,
  # leave out the dest side data
  def sessions_source_preferred
    all_sessions = sessions.all
    result = all_sessions.select do |s|
      s.outbound? || (s.inbound? && !all_sessions.detect { |x| x.session_id == s.session_id && x.outbound?} )
    end
  end

  # return an array of all source paths for all sessions
  # note that due to the fact that source_paths is truncated at 4k, we may get some partial paths
  def all_source_paths
    result = sessions_source_preferred.collect{|s| s.docrooted_source_paths.to_s.split(/[\n\r\t]+/)}.flatten.uniq

    if result.empty?
      # try to get from job_step record instead
      result = self.source_paths.to_s.split(/[\n\r\t]+/).uniq
    end
    result
  end

  def user
    job && job.user
  end
  
  def force_recalc!
    update_attribute(:force_recalc, true)
  end
  
  def aspera_sync?
    job && job.started_via == 'aspera.sync'
  end
  
  def aspera_sync2?
    job && job.started_via == 'aspera.sync2'
  end

  def session_source_address
    session && session.source_address
  end

  def session_dest_address
    session && session.dest_address
  end

  # Returns which FaspSession to use in updating session information on the JobStep
  # row.  If only one node is reporting use its session.  If both are reporting take the one
  # with the larger value in FaspSession#usec.
  def session
    @session ||= if client_session && server_session
        server_session_recent_enough? ? server_session : client_session
      elsif client_session
        client_session
      else
        server_session
      end
  end

  def description
    job || job.description
  end
  
  def label
    nsteps = job.steps.length
    nsteps > 1 ? "#{job.steps.index(self) + 1} of #{nsteps} " : " "
  end

  def status_scheduled?
    status == STATUS_SCHEDULED
  end
  
  def status_queued?
    status == STATUS_QUEUED
  end

  def status_willretry?
    status == STATUS_WILLRETRY
  end

  def status_submitted?
    status == STATUS_SUBMITTED
  end
  
  def status_email_active?
    Console::STATUSES_EMAIL_ACTIVE.include? status
  end
  
  def status_active?
    STATUSES_ACTIVE.include? status
  end
  
  def status_cancelled?
    status == STATUS_CANCELLED
  end
  
  def status_paused?
    status == STATUS_PAUSED
  end

  def status_possibly_active?
    STATUSES_POSSIBLY_ACTIVE.include? status
  end

  def status_possibly_active_and_session_exists?
    ( !session.nil? ) && (STATUSES_POSSIBLY_ACTIVE.include? status)
  end

  def status_running?
    status == STATUS_RUNNING
  end
  
  def running?
    status_running?
  end
  
  def status_success?
    STATUSES_SUCCESS.include? status
  end

  def status_error?
    STATUSES_ERROR.include? status
  end
  
  # Returns +true+ if this JobStep's Job was started via the console.  The :job
  # associated record must be saved and the :job_id column must be set for this 
  # to return the proper value.
  def console?
    job.console?
  end

  # Returns node name if managed or partner node, else ip address.
  def source_node_name
    name_of_node(source_node, source_node_address)
  end
  
  # Returns node name if managed or partner node, else ip address.
  def dest_node_name
    name_of_node(dest_node, dest_node_address)
  end
  
  # Returns true if the source or destination match the node
  def touches_node?(node)
    source_node == node || dest_node == node
  end

  # Calculate a node name:
  # * If _managed_node_ not +nil+ return its +node_name+
  # * If we find a partner node matching _address_, use its +name+
  # * If neither managed nor partner node just return the _address_.
  def name_of_node(managed_node, address)
    if managed_node
      managed_node.node_name
    elsif partner_node = PartnerNode.find_by_address(address)
      partner_node.name || address
    else
      address
    end
  end

  def source_fasp_or_partner_node
    source_node || source_partner_node
  end

  def source_partner_node
    PartnerNode.find_by_address(source_node_address)
  end

  def dest_fasp_or_partner_node
    dest_node || dest_partner_node
  end

  def dest_partner_node
    PartnerNode.find_by_address(dest_node_address)
  end

  def eta
    status_running? ? read_attribute(:eta) : nil
  end
  
  # average rate over entire job step
  def average_rate
    if sessions.empty?
      nil
    else
      bytes = total_bytes_transferred
      seconds = total_time
      seconds = 1 if seconds == 0
      # convert bytes to bits, divide by secs
      (bytes * 8 / seconds).to_i
    end
  rescue
    nil
  end

  def total_files_complete
    sessions_total(:files_complete).to_i - sessions_total(:files_skipped).to_i
  end

  def total_files_failed
    sessions_total(:files_failed)
  end

  def total_files_skipped
    sessions_total(:files_skipped)
  end

  def last_files_skipped
    sessions.last.files_skipped rescue 0
  end

  def unique_session_ids
    sessions.map(&:session_id).uniq
  end

  def multisession?
    unique_session_ids.length > 1
  end

  def total_bytes_transferred
    sessions_total(:bytes_transferred)
  end

  # returns a WHERE condition to retrieve files for a job step
  # note that this will only show part of files for hot folder transfers since they can be divided among multiple job steps
  def file_search_sql(err_only, path_match = nil)
    sessions_str = "(#{self.sessions.map{|s| "'#{s.session_id}'"}.join(',')})"
    session_id_condition = "session_id in #{sessions_str}"

    # we want FaspFile records from only one side if both are managed, so match node_uuid from record chosen
    # by DB Ingest as the 'reference' session for JobStep
    node_condition = FaspFile.public_sanitize_sql_for_conditions(['node_uuid = ?', session && session.node_uuid])
    status_condition = err_only ? "status='error'"  : nil
    path_condition = path_match.present? ? AspExp::Operator[:contain].filter_sql_expression(['file_fullpath', path_match]) : nil
    return [session_id_condition, node_condition, status_condition, path_condition].compact.join(" AND ")
  end

  # get first n files for job_step
  def files(count)
    files_array = []
    file_fullpaths = []
    # decide whether to pull from server side or client side
    if client_sessions.count >= server_sessions.count
      session_list = client_sessions
    else
      session_list = server_sessions
    end
    
    for fasp_session in session_list
      # don't double count files reported by both sides
      for file in fasp_session.files
        unless file_fullpaths.include? file.file_fullpath
          file_fullpaths << file.file_fullpath
          files_array << file
          if files_array.length == count
            return files_array
          end
        end
      end
    end
    files_array
  end

  def elapsed_time
    (ended_at || Time.now) - started_at
  rescue
    nil
  end

  
  def total_time
    seconds_hash = {}
    for fasp_session in sessions
      secs = fasp_session.elapsed_seconds
      # if we already have a value reported from the other side
      if seconds_hash.has_key? fasp_session.session_id
        # don't bother replacing the value from other node if this value is nil
        unless secs.nil?
          # take min of both, protecting against existing val being nil
          if seconds_hash[fasp_session.session_id].nil?
            seconds_hash[fasp_session.session_id] = secs
          else
            seconds_hash[fasp_session.session_id] = [seconds_hash[fasp_session.session_id], secs].min
          end
        end
      else
        seconds_hash[fasp_session.session_id] = secs
      end
    end
    # add them all up, return nil if any are nil
    seconds_total = 0
    seconds_hash.each_value {|v|
      if v.nil?
        return nil
      else
        seconds_total += v
      end
    }
    seconds_total
  end
  
  def total_size
    # FIXME: use actual bytes transferred after step complete
    (session && session.bytes_pretransfer) || 'unknown'
  end

  # used by email notifications
  def files_pretransfer
    session ? session.files_pretransfer : 'unknown'
  end

  def files_count
    session ? session.files_pretransfer : 0
  end

  def bytes_transferred
    session ? session.bytes_transferred : 0
  end

  def priority
    if session
      if session.rates.empty?
        'normal'
      else
        session.rates[-1].priority
      end
    else 
      'normal'
    end
  end

  def session_transport
    session && session.transport
  end
  
  def rate_adjustable?(user)
    session && session.rate_adjustable?(user)
  end
  
  def upload?
    session && session.upload?
  end

  def request_stats_update
    if client_session
      client_session.request_stats_update
    end
    if server_session
      server_session.request_stats_update
    end
  end

  def tags_for_job_request
    return nil if console_tags.blank?

    formatted_tags = {}

    tags = JSON.parse(console_tags)

    tags.each do |k, v|
      formatted_tags[v['name']] = v['value']
    end

    ({'aspera' => {'console' => {'user_specified' => formatted_tags}}})
  end

  def formatted_tags_for_job_request
    tags_for_job_request.to_json
  end

  # sum values for each session, taking max if both sides are reported
  # nil values are converted to zeroes
  def sessions_total(attribute)
    vals = {}
    for fasp_session in sessions
      vals[fasp_session.session_id] = [vals[fasp_session.session_id].to_i, fasp_session[attribute].to_i].max
    end
    vals.values.inject(0) {|total, val| total + val}
  end

  # quick way to see if anything was transferred for this session
  # total can be inaccurate due to double-counting, but can still be used
  # as an indicator that something got transferred.
  def anything_transferred?
    sessions.sum('bytes_transferred') > 0
  end
  
  # Returns +true+ if this step can be inactivated.  If so, an "Inactivate" link is
  # displayed on the session detail page.
  def can_inactivate?
    status == JobStep::STATUS_RUNNING
  end
  
  # Returns the latest FaspSession#db_logger_last_reported_at.
  def db_logger_last_reported_at
    sessions.map(&:db_logger_last_reported_at).max
  end
  
  def inactivate
    if can_inactivate?
      sessions.each(&:manually_mark_inactive)
      self.force_recalc=true
      self.save(false)
    end
  rescue
    logger.error("Exception while trying to mark JobStep as inactive: #{$!} #{$!.backtrace.join("\n")}")
    false
  end

  def should_select_initial_endpoints?
    if !ConsoleConfig::ConsoleDefault.failover_use_original_endpoints_first
      if source_endpoints_tried.nil? && dest_endpoints_tried.nil?
        return true
      end
    end

    return false
  end

  # Before_save callback - select best initial endpoints in failover group
  def select_best_initial_endpoints
    # Determine best source and dest endpoints
    best_src_ep = select_best_endpoint_id(source_saved_login_id, source_endpoints_tried, 'outbound')
    best_dest_ep = select_best_endpoint_id(dest_saved_login_id, dest_endpoints_tried, 'inbound')

    # Update source and dest endpoints if better option is found
    if source_saved_login_id != best_src_ep
      src_login = Login.find_by_id(best_src_ep)

      self.source_saved_login_id = best_src_ep
      self.source_node_address = src_login.address
      self.source_node_id = src_login.owner_id
    end

    # Update source and dest endpoints if better option is found
    if dest_saved_login_id != best_dest_ep
      dest_login = Login.find_by_id(best_dest_ep)

      self.dest_saved_login_id = best_dest_ep
      self.dest_node_address = dest_login.address
      self.dest_node_id = dest_login.owner_id
    end
  end

  # Return the id of the best remaining login (source or destination) in failover group
  # Nil if no options left
  def select_best_endpoint_id(saved_login_id, tried, direction)
    endpoint = Login.find_by_id(saved_login_id)

    return saved_login_id if endpoint.nil?

    # Get endpoints tried as list of ids
    tried = tried.to_s.split(',').map(&:to_i)

    # If not in cluster, return endpoint if not tried yet
    if endpoint.owner.nil? || !endpoint.cluster_node?
      return (tried.include?(endpoint.id) ? nil : endpoint.id)
    end

    alt_endpoints = endpoint.cluster_equivalent_endpoints

    if alt_endpoints.empty?
      # No other options, check if this endpoint has been tried
      if tried.include?(saved_login_id)
        return nil
      else
        return saved_login_id
      end
    end

    # Add original endpoint to front of array
    # If no others are better, will stick with original endpoint
    endpoints = [endpoint] + alt_endpoints

    # Filter out already tried endpoints
    endpoints = endpoints.select{ |ep| !tried.include?(ep.id) }

    return nil if endpoints.empty?

    endpoints.sort! {|x, y| [x.owner.cached_bitrate(direction) || 99999, x.id] <=> [y.owner.cached_bitrate(direction) || 99999, y.id]}

    endpoints[0].id
  end

  # Select the next best source end dest endpoints in failover group
  # Will eventually produce all pairs of endpoints, return nil when no options remain
  def handle_failover
    # Get best endpoint remaining
    src_endpoint_id = select_best_endpoint_id(self.source_saved_login_id, self.source_endpoints_tried, 'outbound')
    dest_endpoint_id = select_best_endpoint_id(self.dest_saved_login_id, self.dest_endpoints_tried, 'inbound')

    if src_endpoint_id.present?
      new_src_login = Login.find_by_id(src_endpoint_id)

      self.attributes = {:source_saved_login_id => src_endpoint_id,
                         :source_node_address => new_src_login.address,
                         :source_node_id => new_src_login.owner_id}

      return save(false)
    elsif dest_endpoint_id.present?
      # Out of source endpoint options - try different dest endpoint
      # and cycle through source endpoints again
      new_dest_login = Login.find_by_id(dest_endpoint_id)

      self.attributes = {:dest_saved_login_id => dest_endpoint_id,
                         :dest_node_address => new_dest_login.address,
                         :dest_node_id => new_dest_login.owner_id}

      # Reset source endpoints tried - will try each source endpoint with each dest endpoint
      self.attributes = {:source_endpoints_tried => nil}

      return save(false)
    end

    logger.error "Error in Job Step #{id}: Failover Possible not set to false when it should have been"

    self.attributes = {:failover_possible => false}
    self.save(false)

    return nil
  end

  # Add the current endpoints to the list of endpoints tried by this job step, if not currently included
  def mark_endpoints_tried
    seps = self.source_endpoints_tried
    src_id = self.source_saved_login_id.to_s

    if seps.nil?
      self.source_endpoints_tried = src_id
    elsif !seps.split(',').include?(src_id)
      self.source_endpoints_tried += ",#{src_id}"
    end

    deps = self.dest_endpoints_tried
    dest_id = self.dest_saved_login_id.to_s

    if deps.nil?
      self.dest_endpoints_tried = dest_id
    elsif !deps.split(',').include?(dest_id)
      self.dest_endpoints_tried += ",#{dest_id}"
    end

    self.save(false)
  end

  # If this step can fail over, mark failover_possible
  def mark_failover_possible
    if can_failover?
      self.failover_possible = true
      self.save(false)
    end
  end

  # Check if there are any remaining nodes to try for failover
  def can_failover?
    select_best_endpoint_id(self.source_saved_login_id, self.source_endpoints_tried, 'outbound') || select_best_endpoint_id(self.dest_saved_login_id, self.dest_endpoints_tried, 'inbound') ? true : false
  end

  class << self

    # pass in a filter expression (e.g. fasp_files.file_basename LIKE 'foo%') and get back an EXISTS clause that can be used
    # as a condition when querying the jobs table
    # N.B.: this is used by Console API call TransferList, and is roughly adapted from Job#get_filename_filter_clause() which
    # serves a similar purpose filtering for the UI Activity page.
    def api_file_filter_clause(filter_sql)
      "EXISTS (SELECT fasp_files.id FROM fasp_files WHERE (fasp_files.session_id IN (select session_id from fasp_sessions where job_step_id = job_steps.id) AND (#{filter_sql})) LIMIT 1)"
    end

    def soap_cleanup(ids)
      ids.each do |step_id|
        step = find(step_id)
        step.attributes = {
          :status   => STATUS_ERROR,
          :err_desc => 'Communication problem with node.  Marked as "error" by admin',
          :ended_at => Time.now,
        }
        step.save(false)
        # update underlying raw data
        # Canonicalizer has its own correction for FaspFile records that are left with status 'running', so correct only FaspSession records
        fasp_sessions = step.sessions.find(:all, :conditions => ["status  = ? OR status = ?", DbLogger::STATUS_STARTED, DbLogger::STATUS_RUNNING])
        fasp_sessions.each do |fs|
          fs.status = DbLogger::STATUS_ERROR
          fs.err_desc = 'Communication problem with node.  Marked as "error" by admin'
          fs.stopped_at = Time.now
          fs.save(false)
        end
      end
      ids.size
    end
    
  end
end
