class JobTemplate < ActiveRecord::Base

  include Common::EarPassword
  include Common::ProxyPassword
  include Common::Obfuscator

  before_save :encrypt_ear_password
  before_save :encrypt_proxy_password

  # non-database attributes to use when saving from form
  attr_accessor :ear_encrypted_password
  attr_accessor :ear_encrypted_password_confirm
  attr_accessor :use_ear

  # FIXME: change this association to be :creator (and remove the #creator method)
  belongs_to :user
  has_many :template_users, :class_name => "JobTemplateUser"
  has_many :users, :through => :template_users

  has_many :jobs
  has_many :steps, :class_name => "JobTemplateStep", :dependent => :destroy

  # FIXME(?): replace job_template_steps references w/steps
  has_many :job_template_steps, :dependent => :destroy

  has_many :emails, :class_name => "JobTemplateEmail", :dependent => :destroy, :order => "locked DESC, email" do
    # Returns template emails sorted: locked, user, other
    def for_start_transfer(user)
      user_jte = JobTemplateEmail.new_from_user(user)  # may be nil
      [locked_for(user), user_jte, not_locked_for(user)].flatten.compact
    end

    # Returns locked JobTemplateEmail's with <tt>current_user</tt>'s removed
    def locked_for(user)
      locked.reject { |jte| jte.email == user.email }
    end

    # Returns not_locked JobTemplateEmail's with <tt>current_user</tt>'s removed
    def not_locked_for(user)
      not_locked.reject { |jte| jte.email == user.email }
    end
  end

  has_one :schedule, :as => :schedulable, :class_name => "ScheduleRule", :dependent => :destroy

  named_scope :shared, :conditions => "shared = '1'"

  validates_presence_of     :user_id

  validates_presence_of     :description,
                            :message => "^Smart transfer name is required"
  validates_length_of       :description,
                            :maximum => 255,
                            :message => "^Smart transfer name must be less than 255 characters"
  validates_uniqueness_of   :description,
                            :message => "^Smart transfer name has already been taken"

  validates_numericality_of :udp_port,
                            :only_integer => true,
                            :allow_nil => true,
                            :greater_than_or_equal_to => 0,
                            :less_than_or_equal_to => 65535

  validates_presence_of     :target_rate
  validates_numericality_of :target_rate,
                            :allow_nil => true, # avoid showing 2 error messages if value is missing
                            :only_integer => true,
                            :greater_than_or_equal_to => 0

  validates_presence_of     :minimum_rate,
                            :unless => Proc.new {|job_template|  job_template.priority_and_policy == 'fixed'}
  validates_numericality_of :minimum_rate,
                            :allow_nil => true, # avoid showing 2 error messages if value is missing
                            :only_integer => true,
                            :greater_than_or_equal_to => 0

  validate                  :validate_minimum_rate_vs_target

  validate                  :validate_end_points
  validate                  :validate_ear_password

  validates_presence_of     :priority_and_policy
  validates_inclusion_of    :priority_and_policy,
                            :in => ['low', 'fair', 'high', 'fixed'],
                            :allow_blank => true # avoid showing 2 error messages if value is unset

  validates_numericality_of :max_retry_attempts,
                            :only_integer => true,
                            :allow_nil => true,
                            :greater_than_or_equal_to => 0

  validates_numericality_of :retry_interval_seconds,
                            :only_integer => true,
                            :allow_nil => true,
                            :greater_than_or_equal_to => 0

  validates_presence_of     :encryption
  validates_inclusion_of    :encryption,
                            :in => ['none', 'aes-128', 'aes-192', 'aes-256'],
                            :allow_blank => true # avoid showing 2 error messages if value is unset

  validates_inclusion_of    :resume_partially_transferred_files, :in => [true, false]

  validates_presence_of     :file_compare_type
  validates_inclusion_of    :file_compare_type,
                            :in => ['Attributes', 'Sparse Checksum', 'Full Checksum'],
                            :allow_blank => true # avoid showing 2 error messages if value is unset

  validates_presence_of     :file_overwrite_policy
  validates_inclusion_of    :file_overwrite_policy,
                            :in => ['Always', 'Diff', 'DiffAndOlder', 'Older', 'Never'],
                            :allow_blank => true # avoid showing 2 error messages if value is unset

  validates_inclusion_of    :preserve_file_timestamps, :in => [true, false]

  validates_numericality_of :datagram_size,
                            :allow_nil => true,
                            :only_integer => true,
                            :greater_than => 0

  validates_numericality_of :read_block_size,
                            :allow_nil => true,
                            :only_integer => true,
                            :greater_than => 0

  validates_numericality_of :write_block_size,
                            :allow_nil => true,
                            :only_integer => true,
                            :greater_than => 0

  validates_presence_of     :file_exclude_newer_than, :if => :file_exclude_by_timestamp
  validates_numericality_of :file_exclude_newer_than, :if => :file_exclude_by_timestamp,
                            :allow_blank => true,
                            :only_integer => false,
                            :greater_than_or_equal_to => 0,
                            :less_than_or_equal_to => 1000000000
  validates_numericality_of :file_exclude_newer_than, :if => :file_exclude_by_timestamp,
                            :only_integer => true,
                            :message => 'must be an integer value'

  validate                 :validate_archive_directory_present
  validate                 :validate_tags

  validate                 :validate_initiator_acl

  validate                 :schedule_validation

  after_destroy :cancel_scheduled_transfers

  def schedule_validation
    schedule.save! if schedule
  end

  def initialize(attributes = {})
    defaults = ConsoleConfig::ConsoleDefault.transfer_defaults
    defaults.delete(:enable_retry)
    super(defaults.merge(attributes))
  end


  def validate_initiator_acl
    steps_for_validation.each do |step|
      step.validate_initiator_acl
    end
  end

  # check to see if template would have a problem if node was converted
  def convert_legacy_acl_problem?(fasp_node)
    steps_for_validation.any? { |step|
      step.convert_legacy_acl_problem?(fasp_node)
    }
  end


  # class methods
  # ============================================================================

  def self.unshare_all_with_personal_logins
    all.each do |job_template|
      if job_template.shared? && job_template.has_personal_login?
        job_template.shared = false
        job_template.save!
      end
    end
  end

  # Process smart transfers that have a schedule
  def self.process_recurring_transfers
    JobTemplate.find_all_by_recurring(true).each do |job_template|
      job_template.reload # reduce window when online change of template not picked up
      job_template.process_recurring_transfers
    end
  end

  # instance methods
  # ============================================================================

  # stub needed to prevent encryption salt from breaking
  def created_at
    return Time.parse('2009-01-01 12:00:00 UTC')
  end

  def start_of_day_in_timezone(date, zone_name)
    ActiveSupport::TimeZone.new(zone_name).parse(date.to_s)
  end

  def currently_scheduled_job
    return nil unless next_runtime

    jobs.scheduled.recurring.find_by_scheduled_start_at(next_runtime)
  end

  def concurrent_instances
    jobs.concurrent
  end

  def next_recurring_transfer_starts_after
    # don't schedule a job for a time before the last cancelled or manually started job
    max_scheduled_at = jobs.not_scheduled.recurring.maximum(:scheduled_start_at) || db_time_now

    [db_time_now, max_scheduled_at].max
  end

  def process_recurring_transfers
    update_attribute(:next_runtime, calculated_next_runtime)
    logger.info "JobTemplate #{id}: next_job_starts_after: #{next_recurring_transfer_starts_after}, next_runtime: #{next_runtime}"
    submit_recurring if next_runtime && currently_scheduled_job.nil?
  end

  # Unlike starting from web page, override_data is much more limited
  # Only contains the following:
  #
  # :description =>
  # :comment =>
  # :source_paths =>
  # :source_base_enable =>
  # :source_base =>
  # :scheduled_start =>
  #
  # Returns a job id
  #
  def submit_from_api_call(initiator, override_data)
    # validate scheduled start if present
    if override_data[:scheduled_start]
      begin
        dt = Time.zone.parse(override_data[:scheduled_start])
        if dt < db_time_now
          errors.add_to_base("Scheduled transfer start must be in the future")
          return false
        end
      rescue
        errors.add_to_base("Scheduled transfer start is not a valid date/time.")
        return false
      end
    end

    # validate source items / source base
    if self.source_type == 'user_selected' && override_data[:source_paths].to_s.strip.blank?
      self.errors.add_to_base("No source paths selected.")
      return false
    end
    
    if self.source_type != 'user_selected' && override_data[:source_paths].to_s.strip.present?
      self.errors.add_to_base("Smart transfer does not allow changing source paths/base.")
      return false
    end

    if self.source_type != 'user_selected' && override_data[:source_base].to_s.strip.present?
      self.errors.add_to_base("Smart transfer does not allow changing source paths/base.")
      return false
    end

    job = Job.new
    job.uuid = UUID.random_create.to_s
    job.description = override_data[:description] || self.description
    job.comment = override_data[:comment].to_s.strip
    job.job_template_id = self.id
    job.started_via = Console::CONSOLE_APPLICATION
    job.user_id = initiator.id
    job.contact = initiator.login + " (console)"

    if override_data[:scheduled_start]
      job.status = Job::STATUS_SCHEDULED
      job.scheduled_start_at = Time.zone.parse(override_data[:scheduled_start])
    else
      job.status = Job::STATUS_QUEUED
    end

    job.recurring = false

    job_template_steps.each do |jts|
      # assemble job step

      js = job.job_steps.build
      js.queueable = override_data[:scheduled_start] ? false : true # transfers initiated by Console are Console-queueable but only when they are due to start
      js.uuid = UUID.random_create.to_s
      if override_data[:scheduled_start]
        js.scheduled_start_at = Time.zone.parse(override_data[:scheduled_start])
        js.status = Job::STATUS_SCHEDULED
      else
        js.status = Job::STATUS_QUEUED
      end
      js.source_node_id = (jts.source_login.fasp_node && jts.source_login.fasp_node.id)
      js.dest_node_id = (jts.dest_login.fasp_node && jts.dest_login.fasp_node.id)
      js.source_initiated = jts.source_initiated

      js.dest_node_address = jts.dest_login.address
      js.dest_path = jts.dest_directory

      js.dest_saved_login_id = jts.dest_saved_login_id
      js.dest_ssh_login = jts.dest_login.login
      js.dest_ssh_port = jts.dest_login.port

      js.source_node_address = jts.source_login.address
      js.source_saved_login_id = jts.source_saved_login_id
      js.source_ssh_login = jts.source_login.login
      js.source_ssh_port = jts.source_login.port

      js.delete_before_transfer = jts.delete_before_transfer || false

      js.archive_after_transfer = jts.archive_after_transfer && self.steps.length == 1
      js.archive_directory = jts.archive_directory
      js.delete_source_after_transfer = jts.delete_source_after_transfer && self.steps.length==1
      js.delete_source_dirs_after_transfer = jts.delete_source_dirs_after_transfer && self.steps.length==1
      js.delete_empty_directories = jts.delete_empty_directories && self.steps.length==1

      # use the individual parameters for this destination set at design time
      js.encryption = jts.encryption
      js.minimum_rate = jts.minimum_rate || 0
      js.target_rate = jts.target_rate

      js.console_tags = jts.console_tags

      js.preserve_file_timestamps = jts.preserve_file_timestamps
      js.preserve_file_create_times = jts.preserve_file_create_times
      js.preserve_file_modify_times = jts.preserve_file_modify_times
      js.preserve_file_access_times = jts.preserve_file_access_times
      js.preserve_file_source_access_times = jts.preserve_file_source_access_times
      js.preserve_file_uids = jts.preserve_file_uids
      js.preserve_file_gids = jts.preserve_file_gids

      js.symbolic_link_handling = jts.symbolic_link_handling
      js.priority_and_policy = jts.priority_and_policy

      js.use_ascp4 = jts.use_ascp4 || false

      js.protocol = (Console.alternate_protocols_enabled ? jts.protocol : 'FASP') || 'FASP'
      js.udp_port = jts.udp_port
      js.use_fasp_proxy = jts.use_fasp_proxy
      js.proxy_host = jts.proxy_host
      js.proxy_port = jts.proxy_port
      js.proxy_secure = jts.proxy_secure
      js.proxy_user = jts.proxy_user
      js.proxy_unencrypted_password = jts.proxy_unencrypted_password

      js.file_exclude_by_timestamp = jts.file_exclude_by_timestamp
      js.file_exclude_newer_than = jts.file_exclude_newer_than

      js.max_retry_attempts = jts.max_retry_attempts
      js.retry_interval_seconds = jts.retry_interval_seconds
      js.file_overwrite_policy = jts.file_overwrite_policy
      js.file_compare_type = jts.file_compare_type
      js.exclude_file_filter = jts.exclude_file_filter
      js.include_file_filter = jts.include_file_filter

      js.prefer_dest_initiated = jts.prefer_dest_initiated

      js.datagram_size = jts.datagram_size
      js.read_block_size = jts.read_block_size
      js.write_block_size = jts.write_block_size
      js.ear_unencrypted_password = jts.ear_unencrypted_password
      js.resume_partially_transferred_files = jts.resume_partially_transferred_files


      # source paths are overrideable if source type = 'user_selected'
      if self.source_type == 'user_selected'
        js.source_paths = override_data[:source_paths]
        js.source_base = override_data[:source_base]
      else
        js.source_paths = jts.source_directory
        js.source_base = jts.source_base_enable ? jts.source_base : nil
      end

      jts.emails.each do |one_email|

        if one_email.start_template_id.to_i > 0 || one_email.success_template_id.to_i > 0 || one_email.error_template_id.to_i > 0
          new_email = js.emails.build
          new_email.email = one_email.email
          new_email.start_template_id = one_email.start_template_id
          new_email.success_template_id = one_email.success_template_id
          new_email.error_template_id = one_email.error_template_id
          new_email.time_zone = JobStepEmail.time_zone_for_email(new_email.email)
        end
      end


      # build emails for computed emails (those due to node endpoint configuration or user preferences)
      js.build_computed_notifications
    end

    if job.save
      job
    else
      # add validation errors to smart transfer base so that they will be displayed by ajax
      job.errors.full_messages.each do |err_msg|
        self.errors.add_to_base err_msg
      end
      job.job_steps.each do |jobstep|
        jobstep.errors.full_messages.each do |err_msg|
          self.errors.add_to_base err_msg
        end
      end
      false
    end
  end


  def submit_from_override_hash(initiator, override_data)
    # perform basic validation before trying anything else
    # description
    if override_data['description'].to_s.strip.blank?
      self.errors.add_to_base("Please fill in a transfer name.")
      return false
    end

    # transfer time
    if override_data["run_when_later"] == "later"
      if override_data["scheduled_start_at"].blank?
        errors.add_to_base("Enter a scheduled start date/time")
        return false
      else
        begin
          dt = Time.zone.parse(override_data["scheduled_start_at"])
          if dt < db_time_now
            errors.add_to_base("Scheduled transfer start must be in the future")
            return false
          end
        rescue
          errors.add_to_base("Scheduled transfer start is not a valid date/time.")
          return false
        end
      end
    end


    # source items
    if override_data['source_type']=='user_selected' && override_data['job_template_steps'][0]['source_paths'].to_s.strip.blank?
      self.errors.add_to_base("Please choose one or more source items.")
      return false
    end

    if override_data['file_exclude_by_timestamp']
      time = override_data['file_exclude_newer_than']

      if time.blank?
        self.errors.add_to_base "File exclude newer can't be blank"
        return false
      elsif time =~ /^\d+\.\d+$/
        self.errors.add_to_base "File exclude newer than must be an integer value"
        return false
      elsif !(time =~ /^\d+$/)
        self.errors.add_to_base "File exclude newer than is not a number"
        return false
      elsif time.to_i < 0
        self.errors.add_to_base "File exclude newer than must be greater than 0"
        return false
      elsif time.to_i > 1000000000
        self.errors.add_to_base "File exclude newer than must be less than 1000000000"
        return false
      end
    end

    # only examine areas of override_data that are allowed to be overridden
    job = Job.new
    job.uuid = UUID.random_create.to_s
    job.description = override_data['description'].to_s.strip
    job.comment = override_data['comment'].to_s.strip
    job.job_template_id = id
    job.started_via = Console::CONSOLE_APPLICATION
    job.user_id = initiator.id
    job.contact = initiator.login + " (console)"

    if override_data['run_when_later']
      job.status = Job::STATUS_SCHEDULED
      job.scheduled_start_at = Time.zone.parse(override_data['scheduled_start_at'])
    else
      job.status = Job::STATUS_QUEUED
    end

    job.recurring = false

    submitted_jts_ids = override_data['job_template_steps'].map{|x| x['id'].to_i}

    job_template_steps.each do |jts|
      # if JobTemplateStep is missing from what was submitted
      if !submitted_jts_ids.include? jts.id.to_i
        # and it was not allowed to be deleted
        unless jts.allow_delete_dest
          # abort
          errors.add_to_base('Invalid smart transfer submission. One or more locked destination nodes were removed from the transfer.')
          return false
        end
      elsif submitted_jts_ids.include?(jts.id.to_i) && jts.both_sides_initiator_problem?
        # there is an initiator problem with this step and it was not deleted
        errors.add_to_base(jts.acl_error_message)
        return false
      else
        # destination was not deleted, assemble job step

        js = job.job_steps.build
        js.queueable = override_data['run_when_later'] ? false : true # transfers initiated by Console are Console-queueable but only when they are due to start
        js.uuid = UUID.random_create.to_s
        if override_data['run_when_later']
          js.scheduled_start_at = Time.zone.parse(override_data['scheduled_start_at'])
          js.status = Job::STATUS_SCHEDULED
        else
          js.status = Job::STATUS_QUEUED
        end
        js.source_node_id = (jts.source_login.fasp_node && jts.source_login.fasp_node.id)
        js.dest_node_id = (jts.dest_login.fasp_node && jts.dest_login.fasp_node.id)
        js.source_initiated = jts.source_initiated

        jts_overrides = override_data['job_template_steps'].find{|x| x['id']==jts.id}

        js.dest_node_address = jts.dest_login.address

        # destination path - if it is allowed to be overridden, use whatever was in the override
        if jts.allow_change_dest_dir
          js.dest_path = jts_overrides['dest_directory']
        else
          js.dest_path = jts.dest_directory
        end

        js.dest_saved_login_id = jts.dest_saved_login_id
        js.dest_ssh_login = jts.dest_login.login
        js.dest_ssh_port = jts.dest_login.port

        js.source_node_address = jts.source_login.address
        js.source_saved_login_id = jts.source_saved_login_id
        js.source_ssh_login = jts.source_login.login
        js.source_ssh_port = jts.source_login.port

        js.console_tags = override_data['console_tags'].to_json.to_s if override_data['console_tags'].present?

        if self.allow_initiator_changes
          # if this destination was individually overridden
          if jts_overrides['modify_dest_options_for_start'] && submitted_jts_ids.length > 1
            # take JobStep-level overrides
            # multi-step jobs can't have 'delete source after transfer' set true
            js.delete_before_transfer = jts_overrides['delete_before_transfer']
            js.archive_after_transfer = jts_overrides['archive_after_transfer'] == true && submitted_jts_ids.length == 1
            js.archive_directory = jts_overrides['archive_directory']
            js.delete_source_after_transfer = ((jts_overrides['delete_source_after_transfer']==true) && (submitted_jts_ids.length == 1))
            js.delete_source_dirs_after_transfer = ((jts_overrides['delete_source_dirs_after_transfer']==true) && (submitted_jts_ids.length == 1))
            js.delete_empty_directories = ((jts_overrides['delete_empty_directories']==true) && (submitted_jts_ids.length == 1))
            js.encryption = jts_overrides['encryption']
            js.minimum_rate = jts_overrides['minimum_rate'] || 0
            js.target_rate = jts_overrides['target_rate']

            js.preserve_file_timestamps = jts_overrides['preserve_file_timestamps']
            js.preserve_file_create_times = jts_overrides['preserve_file_create_times']
            js.preserve_file_modify_times = jts_overrides['preserve_file_modify_times']
            js.preserve_file_access_times = jts_overrides['preserve_file_access_times']
            js.preserve_file_source_access_times = jts_overrides['preserve_file_source_access_times']
            js.preserve_file_uids = jts_overrides['preserve_file_uids']
            js.preserve_file_gids = jts_overrides['preserve_file_gids']

            js.symbolic_link_handling = jts_overrides['symbolic_link_handling']
            js.priority_and_policy = jts_overrides['priority_and_policy']

            js.use_ascp4 = jts_overrides['use_ascp4'] || false

            js.protocol = (Console.alternate_protocols_enabled ? jts_overrides['protocol'] : 'FASP') || 'FASP'
            js.udp_port = jts_overrides['udp_port']

            # proxy fields cannot be overridden
            js.use_fasp_proxy = jts.use_fasp_proxy
            js.proxy_host = jts.proxy_host
            js.proxy_port = jts.proxy_port
            js.proxy_secure = jts.proxy_secure
            js.proxy_user = jts.proxy_user
            js.proxy_unencrypted_password = jts.proxy_unencrypted_password

            js.file_exclude_by_timestamp = jts_overrides['file_exclude_by_timestamp']
            js.file_exclude_newer_than = jts_overrides['file_exclude_newer_than']

            js.max_retry_attempts = jts_overrides['max_retry_attempts']
            js.retry_interval_seconds = jts_overrides['retry_interval_seconds']
            js.file_overwrite_policy = jts_overrides['file_overwrite_policy']
            js.file_compare_type = jts_overrides['file_compare_type']
            js.exclude_file_filter = jts_overrides['exclude_file_filter']
            js.include_file_filter = jts_overrides['include_file_filter']

            js.prefer_dest_initiated = jts_overrides['prefer_dest_initiated']

            js.datagram_size = jts_overrides['datagram_size']
            js.read_block_size = jts_overrides['read_block_size']
            js.write_block_size = jts_overrides['write_block_size']
            # if this job template step had a password configured already, check to see if it was overridden
            # if it doesn't have a password already, see if one was set
            if jts.has_ear_password? && jts_overrides['override_ear'] || !jts.has_ear_password?
              # use override
              if jts_overrides['use_ear']
                if jts_overrides['ear_encrypted_password'] != jts_overrides['ear_encrypted_password_confirm']
                  errors.add_to_base('Encryption password does not match confirmation.')
                  return false
                elsif jts_overrides['ear_encrypted_password'].blank?
                  errors.add_to_base('Content protection set but no password provided.')
                  return false
                else
                  js.ear_unencrypted_password = jts_overrides['ear_encrypted_password']
                end
              end
            else
              # use presupplied
              js.ear_unencrypted_password = jts.ear_unencrypted_password
            end

            js.resume_partially_transferred_files = (jts_overrides['resume_partially_transferred_files']==true)
          else
            # take presets
            # (except for delete_source_after_transfer and delete_source_dirs_after_transfer)
            if submitted_jts_ids.length==1
              js.archive_after_transfer = jts_overrides['archive_after_transfer'] == true
              js.archive_directory = jts_overrides['archive_directory']
              js.delete_source_after_transfer = (jts_overrides['delete_source_after_transfer']==true)
              js.delete_source_dirs_after_transfer = (jts_overrides['delete_source_dirs_after_transfer']==true)
              js.delete_empty_directories = (jts_overrides['delete_empty_directories']==true)
            else
              # multi-step jobs can't have 'delete source after transfer' set true
              js.archive_after_transfer = false
              js.delete_source_after_transfer = false
              js.delete_source_dirs_after_transfer = false
              js.delete_empty_directories = false
            end

            # if this destination did not have settings individually set at design time
            # then use job level overrides
            # TODO: interface needs to be revised to make clear at start time which destinations
            # have individually set parameters
            if !jts.override_template_defaults || submitted_jts_ids.length==1
              js.delete_before_transfer = override_data['delete_before_transfer'] || false

              js.encryption = override_data['encryption']
              js.minimum_rate = override_data['minimum_rate'] || 0
              js.target_rate = override_data['target_rate']

              js.preserve_file_timestamps = override_data['preserve_file_timestamps']
              js.preserve_file_create_times = override_data['preserve_file_create_times']
              js.preserve_file_modify_times = override_data['preserve_file_modify_times']
              js.preserve_file_access_times = override_data['preserve_file_access_times']
              js.preserve_file_source_access_times = override_data['preserve_file_source_access_times']
              js.preserve_file_uids = override_data['preserve_file_uids']
              js.preserve_file_gids = override_data['preserve_file_gids']

              js.symbolic_link_handling = override_data['symbolic_link_handling']
              js.priority_and_policy = override_data['priority_and_policy']

              js.use_ascp4 = override_data['use_ascp4'] || false

              js.protocol = (Console.alternate_protocols_enabled ? override_data['protocol'] : 'FASP') || 'FASP'
              js.udp_port = override_data['udp_port']

              # proxy fields cannot be overridden
              js.use_fasp_proxy = use_fasp_proxy
              js.proxy_host = proxy_host
              js.proxy_port = proxy_port
              js.proxy_secure = proxy_secure
              js.proxy_user = proxy_user
              js.proxy_unencrypted_password = proxy_unencrypted_password

              js.file_exclude_by_timestamp = override_data['file_exclude_by_timestamp']
              js.file_exclude_newer_than = override_data['file_exclude_newer_than']

              js.max_retry_attempts = override_data['max_retry_attempts']
              js.retry_interval_seconds = override_data['retry_interval_seconds']
              js.file_overwrite_policy = override_data['file_overwrite_policy']
              js.file_compare_type = override_data['file_compare_type']
              js.exclude_file_filter = override_data['exclude_file_filter']
              js.include_file_filter = override_data['include_file_filter']

              js.prefer_dest_initiated = override_data['prefer_dest_initiated']

              js.datagram_size = override_data['datagram_size']
              js.read_block_size = override_data['read_block_size']
              js.write_block_size = override_data['write_block_size']

              # if this job template step had a password configured already, check to see if it was overridden
              # if it doesn't have a password already, see if one was set
              if has_ear_password? && override_data['override_ear'] || !has_ear_password?
                # use override
                if override_data['use_ear']
                  if override_data['ear_encrypted_password'] != override_data['ear_encrypted_password_confirm']
                    errors.add_to_base('Encryption password does not match confirmation.')
                    return false
                  elsif override_data['ear_encrypted_password'].blank?
                    errors.add_to_base('Content protection set but no password provided.')
                    return false
                  else
                    js.ear_unencrypted_password = override_data['ear_encrypted_password']
                  end
                end
              else
                # use presupplied one set at design time
                js.ear_unencrypted_password = ear_unencrypted_password
              end
              js.resume_partially_transferred_files = (override_data['resume_partially_transferred_files']==true)
            else
              js.delete_before_transfer = jts.delete_before_transfer || false
              # use the individual parameters for this destination set at design time
              js.encryption = jts.encryption
              js.minimum_rate = jts.minimum_rate || 0
              js.target_rate = jts.target_rate

              js.preserve_file_timestamps = jts.preserve_file_timestamps
              js.preserve_file_create_times = jts.preserve_file_create_times
              js.preserve_file_modify_times = jts.preserve_file_modify_times
              js.preserve_file_access_times = jts.preserve_file_access_times
              js.preserve_file_source_access_times = jts.preserve_file_source_access_times
              js.preserve_file_uids = jts.preserve_file_uids
              js.preserve_file_gids = jts.preserve_file_gids

              js.symbolic_link_handling = jts.symbolic_link_handling
              js.priority_and_policy = jts.priority_and_policy

              js.protocol = (Console.alternate_protocols_enabled ? jts.protocol : 'FASP') || 'FASP'

              js.use_ascp4 = jts.use_asp4

              js.udp_port = jts.udp_port
              js.use_fasp_proxy= jts.use_fasp_proxy
              js.proxy_host = jts.proxy_host
              js.proxy_port = jts.proxy_port
              js.proxy_secure = jts.proxy_secure
              js.proxy_user = jts.proxy_user
              js.proxy_unencrypted_password = jts.proxy_unencrypted_password

              js.file_exclude_by_timestamp = jts.file_exclude_by_timestamp
              js.file_exclude_newer_than = jts.file_exclude_newer_than

              js.max_retry_attempts = jts.max_retry_attempts
              js.retry_interval_seconds = jts.retry_interval_seconds
              js.file_overwrite_policy = jts.file_overwrite_policy
              js.file_compare_type = jts.file_compare_type
              js.exclude_file_filter = jts.exclude_file_filter
              js.include_file_filter = jts.include_file_filter

              js.prefer_dest_initiated = jts.prefer_dest_initiated

              js.datagram_size = jts.datagram_size
              js.read_block_size = jts.read_block_size
              js.write_block_size = jts.write_block_size
              js.ear_unencrypted_password = jts.ear_unencrypted_password
              js.resume_partially_transferred_files = jts.resume_partially_transferred_files
            end
          end
        else
          js.delete_before_transfer = jts.delete_before_transfer || false
          js.archive_after_transfer = jts.archive_after_transfer && submitted_jts_ids.length == 1
          js.archive_directory = jts.archive_directory
          js.delete_source_after_transfer = jts.delete_source_after_transfer && submitted_jts_ids.length==1
          js.delete_source_dirs_after_transfer = jts.delete_source_dirs_after_transfer && submitted_jts_ids.length==1
          js.delete_empty_directories = jts.delete_empty_directories && submitted_jts_ids.length==1
          js.encryption = jts.encryption
          js.minimum_rate = jts.minimum_rate || 0
          js.target_rate = jts.target_rate

          js.preserve_file_timestamps = jts.preserve_file_timestamps
          js.preserve_file_create_times = jts.preserve_file_create_times
          js.preserve_file_modify_times = jts.preserve_file_modify_times
          js.preserve_file_access_times = jts.preserve_file_access_times
          js.preserve_file_source_access_times = jts.preserve_file_source_access_times
          js.preserve_file_uids = jts.preserve_file_uids
          js.preserve_file_gids = jts.preserve_file_gids

          js.symbolic_link_handling = jts.symbolic_link_handling
          js.priority_and_policy = jts.priority_and_policy

          js.use_ascp4 = jts.use_ascp4 || false

          js.protocol = (Console.alternate_protocols_enabled ? jts.protocol : 'FASP') || 'FASP'

          js.udp_port = jts.udp_port
          js.use_fasp_proxy= jts.use_fasp_proxy
          js.proxy_host = jts.proxy_host
          js.proxy_port = jts.proxy_port
          js.proxy_secure = jts.proxy_secure
          js.proxy_user = jts.proxy_user
          js.proxy_unencrypted_password = jts.proxy_unencrypted_password

          js.file_exclude_by_timestamp = jts.file_exclude_by_timestamp
          js.file_exclude_newer_than = jts.file_exclude_newer_than

          js.max_retry_attempts = jts.max_retry_attempts
          js.retry_interval_seconds = jts.retry_interval_seconds
          js.file_overwrite_policy = jts.file_overwrite_policy
          js.file_compare_type = jts.file_compare_type
          js.exclude_file_filter = jts.exclude_file_filter
          js.include_file_filter = jts.include_file_filter

          js.prefer_dest_initiated = jts.prefer_dest_initiated

          js.datagram_size = jts.datagram_size
          js.read_block_size = jts.read_block_size
          js.write_block_size = jts.write_block_size
          js.ear_unencrypted_password = jts.ear_unencrypted_password
          js.resume_partially_transferred_files = jts.resume_partially_transferred_files
        end

        # check to see if we need to change source_initiated
        if js.prefer_dest_initiated != jts.prefer_dest_initiated
          js.source_initiated = Login::choose_source_initiated(js.source_login, js.dest_login, js.prefer_dest_initiated)
        end

        # source paths are overrideable if source type = 'user_selected'
        if self.source_type == 'user_selected'
          js.source_paths = override_data['job_template_steps'][0]['source_paths']
          js.source_base = override_data['job_template_steps'][0]['source_base_enable'] ? override_data['job_template_steps'][0]['source_base'] : nil

          # Validate source paths
          # Split and strip
          paths = js.source_paths.split("\n").collect{|p| p.strip.gsub(/\/+$/, "")}

          if paths.detect{ |e| paths.count(e) > 1 }.present?
            errors.add_to_base("Source paths must be unique")
            return false
          end
        else
          js.source_paths = jts.source_directory
          js.source_base = jts.source_base_enable ? jts.source_base : nil
        end

        # emails are always overrideable
        if jts_overrides['override_template_defaults']
          jts_overrides['emails'].each do |email_hash|
            # only build the email if at least one template has been selected
            if email_hash['start_template_id'].to_i > 0 || email_hash['success_template_id'].to_i > 0 || email_hash['error_template_id'].to_i > 0
              new_email = js.emails.build
              new_email.email = email_hash['email']
              new_email.start_template_id = email_hash['start_template_id'].to_i
              new_email.success_template_id = email_hash['success_template_id'].to_i
              new_email.error_template_id = email_hash['error_template_id'].to_i
              new_email.time_zone = JobStepEmail.time_zone_for_email(new_email.email)
            end
          end
        else
          override_data['emails'].each do |email_hash|
            if email_hash['start_template_id'].to_i > 0 || email_hash['success_template_id'].to_i > 0 || email_hash['error_template_id'].to_i > 0
              new_email = js.emails.build
              new_email.email = email_hash['email']
              new_email.start_template_id = email_hash['start_template_id'].to_i
              new_email.success_template_id = email_hash['success_template_id'].to_i
              new_email.error_template_id = email_hash['error_template_id'].to_i
              new_email.time_zone = JobStepEmail.time_zone_for_email(new_email.email)
            end
          end
        end

        # build emails for computed emails (those due to node endpoint configuration or user preferences)
        js.build_computed_notifications
      end
    end
    if job.save
      true
    else
      # add validation errors to smart transfer base so that they will be displayed by ajax
      job.errors.full_messages.each do |err_msg|
        self.errors.add_to_base err_msg
      end
      job.job_steps.each do |jobstep|
        jobstep.errors.full_messages.each do |err_msg|
          self.errors.add_to_base err_msg
        end
      end
      false
    end
  end

  def parsed_tags
    console_tags.blank? ? {} : JSON.parse(console_tags)
  end

  def validate_tags
    return true if console_tags.nil?

    names = {}

    # Check for blank names and values, and ensure uniqueness of tag names
    JSON.parse(console_tags).each do |k, v|
      errors.add_to_base("All tags must have a name") if v['name'].blank?
      errors.add_to_base("All tags must have a value") if v['value'].blank?

      errors.add_to_base("Tag names must be unique (Duplicate tag name: #{v['name']})") if names.has_key?(v['name'])

      names[v['name']] = v['value']
    end
  end

  def validate_archive_directory_present
    job_template_steps.each do |step|
      if step.archive_after_transfer?
        errors.add(:archive_directory, "must be specified if archive after transfer is selected") unless step.archive_directory.present?
      end
    end
  end

  def validate_ear_password
    if ear_encrypted_password != ear_encrypted_password_confirm
      errors.add_to_base("Encryption password does not match confirmation")
      return false
    elsif use_ear && ear_encrypted_password.blank?
      errors.add_to_base("File encryption selected but no password supplied")
      return false
    end
  end

  def validate_and_assign_date(field, value, options)
    options.merge({:allow_nil => false})
    if value =~ /^\s*\"\s*\"\s*$/ # like '""' or with any number of spaces: ' "  " '
      value = nil
    end
    if options[:allow_nil] && value.blank?
      send("#{field}=", nil)
    elsif value.blank?
      errors.add(field, "date must not be blank.")
    else
      begin
        date = Time.zone.parse(value)
        if date.nil?
          errors.add(field, "invalid date.")
        else
          send("#{field}=", date)
        end
      rescue ArgumentError
        errors.add(field, $!.to_s)
      end
    end
  end

  def has_ear_password?
    ear_unencrypted_password?
  end

  def json_string_for_edit()
    self.to_json(
        :include => {
            :job_template_steps => {
                :include => {
                    :emails => {},
                },
                :methods => [:has_ear_password?, :ear_unencrypted_password, :has_proxy_password?, :proxy_unencrypted_password],
            },
            :emails => {},
        },
        :methods => [:has_ear_password?, :ear_unencrypted_password, :has_proxy_password?, :proxy_unencrypted_password]
    )
  end

  def json_string_for_start()
    # need to leave out EAR password & proxy password, just indicate whether a password is set
    self.to_json(
        :include => {
            :job_template_steps => {
                :include => {
                    :emails => {},
                },
                :methods => [:has_ear_password?, :modify_dest_options_for_start, :has_proxy_password?],
            },
            :emails => {},
        },
        :methods => [:has_ear_password?]
    )
  end

  def validate_minimum_rate_vs_target
    if minimum_rate.to_i > target_rate.to_i
      self.errors.add_to_base("Minimum rate cannot be higher than target rate.")
    end
  end

  def start_type
    if source_type == 'incoming_aspera'
      "Triggered by incoming Aspera transfers"
    elsif source_type == 'incoming_any'
      "Triggered by incoming files"
    elsif recurring
      "Repeating: #{schedule_desc}"
    else
      "On request"
    end
  end

  # schedule a job
  def submit_recurring
    logger.info "JobTemplate #{id}: submit_recurring"
    job = Job.new
    job.uuid = UUID.random_create.to_s
    job.description = description
    job.job_template_id = id
    job.started_via = Console::CONSOLE_APPLICATION
    job.status = Job::STATUS_SCHEDULED
    job.scheduled_start_at = next_runtime
    job.user_id = user_id
    job.contact = contact_user + " (console)"
    job.recurring = true
    job_template_steps.each do |jts|
      js = job.job_steps.build
      js.uuid = UUID.random_create.to_s
      js.status = Job::STATUS_SCHEDULED
      js.source_node_id = (jts.source_login.fasp_node && jts.source_login.fasp_node.id)
      js.dest_node_id = (jts.dest_login.fasp_node && jts.dest_login.fasp_node.id)
      js.source_initiated = jts.source_initiated

      js.dest_node_address = jts.dest_login.address
      js.delete_before_transfer = jts.delete_before_transfer
      js.archive_after_transfer = jts.archive_after_transfer
      js.archive_directory = jts.archive_directory
      js.delete_source_after_transfer = jts.delete_source_after_transfer
      js.delete_source_dirs_after_transfer = jts.delete_source_dirs_after_transfer
      js.delete_empty_directories = jts.delete_empty_directories
      js.dest_path = jts.dest_directory
      js.dest_saved_login_id = jts.dest_saved_login_id
      js.dest_ssh_login = jts.dest_login.login
      js.dest_ssh_port = jts.dest_login.port
      js.encryption = jts.encryption
      js.minimum_rate = jts.minimum_rate || 0
      js.source_node_address = jts.source_login.address
      js.source_paths = jts.source_directory
      js.source_saved_login_id = jts.source_saved_login_id
      js.source_ssh_login = jts.source_login.login
      js.source_ssh_port = jts.source_login.port
      js.target_rate = jts.target_rate
      js.scheduled_start_at = next_runtime

      js.console_tags = jts.console_tags

      js.preserve_file_timestamps = jts.preserve_file_timestamps
      js.preserve_file_create_times = jts.preserve_file_create_times
      js.preserve_file_modify_times = jts.preserve_file_modify_times
      js.preserve_file_access_times = jts.preserve_file_access_times
      js.preserve_file_source_access_times = jts.preserve_file_source_access_times
      js.preserve_file_uids = jts.preserve_file_uids
      js.preserve_file_gids = jts.preserve_file_gids

      js.symbolic_link_handling = jts.symbolic_link_handling
      js.priority_and_policy = jts.priority_and_policy

      js.use_ascp4 = jts.use_ascp4 || false

      js.protocol = (Console.alternate_protocols_enabled ? jts.protocol : 'FASP') || 'FASP'

      js.udp_port = jts.udp_port

      js.source_base = jts.source_base_enable ? jts.source_base : nil

      js.use_fasp_proxy = jts.use_fasp_proxy
      js.proxy_host = jts.proxy_host
      js.proxy_port = jts.proxy_port
      js.proxy_secure = jts.proxy_secure
      js.proxy_user = jts.proxy_user
      js.proxy_unencrypted_password = jts.proxy_unencrypted_password

      js.file_exclude_by_timestamp = jts.file_exclude_by_timestamp
      js.file_exclude_newer_than = jts.file_exclude_newer_than

      js.max_retry_attempts = jts.max_retry_attempts
      js.retry_interval_seconds = jts.retry_interval_seconds
      js.file_overwrite_policy = jts.file_overwrite_policy
      js.file_compare_type = jts.file_compare_type
      js.exclude_file_filter = jts.exclude_file_filter
      js.include_file_filter = jts.include_file_filter

      js.prefer_dest_initiated = jts.prefer_dest_initiated

      js.datagram_size = jts.datagram_size
      js.read_block_size = jts.read_block_size
      js.write_block_size = jts.write_block_size
      js.ear_unencrypted_password= jts.ear_unencrypted_password
      js.resume_partially_transferred_files = jts.resume_partially_transferred_files

      jts.emails.each do |jte|
        new_email = js.emails.build
        new_email.email = jte.email
        new_email.start_template_id = jte.start_template_id
        new_email.success_template_id = jte.success_template_id
        new_email.error_template_id = jte.error_template_id
        new_email.time_zone = JobStepEmail.time_zone_for_email(new_email.email)
      end
    end
    job.save!
  end

  def contact_user
    u = User.find_by_id(user_id)
    if u.present?
      u.login
    else
      "deleted user"
    end
  end

  def schedule_desc
    if schedule
      str = schedule.humanize
      str += "<br />Next run: #{next_runtime.blank? ? 'none' : next_runtime_s}"
      str
    end
  end

  def next_runtime_s
    time_in_zone = next_runtime.in_time_zone(schedule.time_zone)
    time_in_zone.to_s(:long_ordinal) + " (#{time_in_zone.zone})"
  end

  def calculated_next_runtime
    return nil unless schedule

    after_time = next_recurring_transfer_starts_after.to_time
    schedule.next_occurrence(after_time)
  end

  # on the assumption that smart transfers are always one-source-to-many-destinations,
  # we return the single source login by examining the first step
  # (assumes that job_template_steps for a job_template are saved in an order
  #  that guarantees that the job_template_step with the earliest source is saved first,
  #  and that the job_template_steps collection is in creation order)
  def source_login
    steps.first && steps.first.source_login
  end

  # This check was in two places in transfer_controller (now three) but when can it ever happen?
  # maybe when someone tries to save a transfer without specifying a source?
  def corrupted?
    source_login.nil? # only happens if template is corrupt
  end

  def source_directory
    steps.first && steps.first.source_directory
  end

  def creator
    user
  end

  def created_by?(current_user)
    user_id == current_user.id
  end

  def created_by_an_admin?
    User.find(user_id).admin?
  rescue
    # user may have been deleted
    false
  end

  def visible_by?(current_user)
    created_by?(current_user) ||
        current_user.admin? ||
        startable_by?(current_user)
  end

  def has_personal_login?
    steps.any? do |step|
      step.source_login.personal? || step.dest_login.personal?
    end
  end

  # if there are steps, iterate through them,
  # test that the login is from a USER, if so, the login should not equal current_user
  # NOTE: the test actually tests whether the current_user DOESNT have the source/dest_logins information
  # and returns TRUE if it doesn't, FALSE if it does
  def has_other_users_personal_login?(current_user)
    steps.any? do |step|
      (step.source_login.personal? && step.source_login.owner != current_user) ||
          (step.dest_login.personal? && step.dest_login.owner != current_user)
    end
  end

  def startable_by?(current_user)
    if current_user.admin?
      startable_by_admin?(current_user)
    else
      startable_by_non_admin?(current_user)
    end
  end

  def startable_by_admin?(current_user)
    return true if created_by?(current_user)
    return false if !console_config_console_default.allow_sharing_personal_smart_transfers && has_other_users_personal_login?(current_user)
    shared?
  end

  def startable_by_non_admin?(current_user)
    if steps_within_allowed_transfer_paths_for_start?(current_user)
      return true if created_by?(current_user)
      return false if !console_config_console_default.allow_sharing_personal_smart_transfers && has_other_users_personal_login?(current_user)
      shared?
    else
      false
    end
  end

  def editable_by_non_admin?(current_user)
    if steps_within_allowed_transfer_paths_for_create?(current_user)
      return true if created_by?(current_user)
      return false if !console_config_console_default.allow_sharing_personal_smart_transfers && has_other_users_personal_login?(current_user)
      shared?
    else
      false
    end
  end

  def editable_by?(current_user)
    if current_user.admin?
      created_by?(current_user) || (ConsoleConfig::ConsoleDefault.load_settings.admins_edit_other_transfers? && created_by_an_admin? && !has_personal_login?)
    else
      created_by?(current_user) && editable_by_non_admin?(current_user)
    end
  end

  def copyable_by?(current_user)
    editable_by?(current_user)
  end

  def deletable_by?(current_user)
    current_user.admin? || created_by?(current_user)
  end

  def console_config_console_default
    @console_default ||= ConsoleConfig::ConsoleDefault.load_settings
  end

  def steps_within_allowed_transfer_paths_for_start?(current_user)
    steps.all? do |step|
      current_user.transfer_paths_for_start_smart.any? do |path|
        path.include?(step.transfer_path) && passes_path_wildcard_tests_for_start?(current_user, path, step)
      end
    end
  end

  def steps_within_allowed_transfer_paths_for_create?(current_user)
    steps.all? do |step|
      current_user.transfer_paths_for_create_smart.any? do |path|
        step_within_transfer_path_and_all_personal_logins_owned_by?(current_user, path, step)
      end
    end
  end

  def step_within_transfer_path_and_all_personal_logins_owned_by?(current_user, path, step)
    path.include?(step.transfer_path) && (path.source_end_point.wildcards? && step.source_login.owner == current_user || !path.source_end_point.wildcards?) &&
        (path.dest_end_point.wildcards?   && step.dest_login.owner == current_user   || !path.dest_end_point.wildcards?)
  end


  def passes_path_wildcard_tests_for_start?(user, path, step)
    passes_path_source_wildcard_tests_for_start?(user, path, step) && passes_path_dest_wildcard_tests_for_start?(user, path, step)
  end

  def passes_path_source_wildcard_tests_for_start?(user, path, step)
    passes_endpoint_wildcard_tests_for_start?(user, path.source_end_point, step.source_login)
  end

  def passes_path_dest_wildcard_tests_for_start?(user, path, step)
    passes_endpoint_wildcard_tests_for_start?(user, path.dest_end_point, step.dest_login)
  end

  def passes_endpoint_wildcard_tests_for_start?(user, end_point, login)
    if end_point.wildcards?
      case login.owner_type
        when 'User'
          login.owner == user || ConsoleConfig::ConsoleDefault.allow_sharing_personal_smart_transfers
        when 'PartnerNode', 'FaspNode', nil
          ConsoleConfig::ConsoleDefault.allow_starting_smart_transfers_matching_any
        else
          # should not find a Smart Transfer with any other type of login owner
          false
      end
    else
      # path endpoint has no wildcards
      true
    end
  end

  def <=>(other)
    description <=> other.description
  end

  def deep_clone current_user
    smart_transfer = self.clone
    smart_transfer.user_id = current_user.id
    smart_transfer.shared = false
    smart_transfer.description = self.unique_name_for_copy
    # Fix #6300 by copying the ear_password as well
    smart_transfer.use_ear                        = self.use_ear
    smart_transfer.ear_unencrypted_password       = self.ear_unencrypted_password
    # Do what the Javascript in smart.html.haml does - use the unencrypted value 3X
    # so we can pass validation.
    smart_transfer.ear_encrypted_password         = self.ear_unencrypted_password
    smart_transfer.ear_encrypted_password_confirm = self.ear_unencrypted_password

    smart_transfer.proxy_unencrypted_password     = self.proxy_unencrypted_password

    smart_transfer.file_exclude_by_timestamp      = self.file_exclude_by_timestamp
    smart_transfer.file_exclude_newer_than      = self.file_exclude_newer_than

    smart_transfer.emails = emails.collect { |e| e.clone }
    smart_transfer.steps  = steps.collect {|s|
      c = s.clone
      c.ear_unencrypted_password = s.ear_unencrypted_password
      c.proxy_unencrypted_password = s.proxy_unencrypted_password
      c
    }

    if self.schedule
      smart_transfer.schedule = self.schedule.clone
      smart_transfer.next_runtime = smart_transfer.calculated_next_runtime
    end

    return smart_transfer
  end

  # If name is "Foo", returns names like "Copy of Foo", "Copy 2 of Foo", etc.
  def unique_name_for_copy
    base_name = description.sub(/^Copy.*?of /, '')
    ([''] + (2..99).to_a).each do |counter|
      new_name = "Copy #{counter} of #{base_name}".squish
      return new_name unless self.class.find_by_description(new_name)
    end
    # We really never expect to get here but in case we do put out a unique name
    return "Copy #{rand} of #{base_name}".squish
  end

  def update_and_save_from_hash(user, data)
    # use a transaction to wrap the validation of user edit
    # if errors are found we raise an exception to trigger the restoration of pre-edit value
    begin
      transaction do
        steps.destroy_all
        emails.destroy_all
        destroy_scheduled_recurring_jobs
        jobs.recurring.update_all('scheduled_start_at = NULL')
        parse_hash_data(user, data)
        self.save!
      end
      true
    rescue   Exception
      # log the exception in case it was not a simple edit validation error
      logger.error "\n\n>>>>>error parsing template data: #{$!}: #{$!.class.name}\n#{$!.backtrace.join("\n")}"
      false
    end
  end

  def destroy_scheduled_recurring_jobs
    jobs.scheduled.recurring.destroy_all
  end

  # Creates and saves a smart transfer
  def new_from_hash(owner, data)
    parse_hash_data(owner, data)
  end

  def parse_hash_data(owner, data)
    # self.data_from_json = data
    update_main_data(owner, data)
    update_step_data(owner, data)
    prevent_template_default_validation_errors if all_dests_individually_set?
  end

  # do all job_template_steps have their settings individually set?
  def all_dests_individually_set?
    !job_template_steps.detect {|s| !s.override_template_defaults}
  end

  def steps_for_validation
    # depending on which association was used during creation of this job_template, one
    # of these will have the steps, the other will (hopefully) be empty.
    job_template_steps.any? ? job_template_steps : steps
  end

  def validate_end_points
    if steps_for_validation.empty?
      errors.add_to_base("A Smart Transfer requires at least one source and destination.")
      return
    end

    source_node = nil


    if steps_for_validation.present? && steps_for_validation.first
      source_node = steps_for_validation.first.source_node
    end

    if use_ear
      if source_node.nil?
        errors.add(:encrypt_at_rest_enable, "^Content protection only supported if the transfer source is a managed node")
      elsif !source_node.supports_encryption_at_rest?
        errors.add(:encrypt_at_rest_enable,
                   "^Content protection requires node to be running 2.2 or higher ")
      end
    end
  end

  # prevent job_template-level validation errors for transfer settings by copying from
  # first step
  def prevent_template_default_validation_errors
    step = job_template_steps[0]

    self.target_rate = step.target_rate
    self.minimum_rate = step.minimum_rate
    #checkbox/multiple-choice fields left out since they would never be missing
    #step.resume_partially_transferred_files
    #step.preserve_file_timestamps
    #step.priority_and_policy

    self.use_ascp4 = step.use_ascp4

    #optional fields would not matter if missing, but could still be set to invalid value, so copy these
    self.protocol = (Console.alternate_protocols_enabled ? step.protocol : 'FASP') || 'FASP'
    self.udp_port = step.udp_port
    self.use_fasp_proxy = step.use_fasp_proxy
    self.proxy_host = step.proxy_host
    self.proxy_port = step.proxy_port
    self.proxy_secure = step.proxy_secure
    self.proxy_user = step.proxy_user
    self.proxy_unencrypted_password = step.proxy_unencrypted_password

    self.file_exclude_by_timestamp = step.file_exclude_by_timestamp
    self.file_exclude_newer_than = step.file_exclude_newer_than

    self.max_retry_attempts = step.max_retry_attempts
    self.retry_interval_seconds = step.retry_interval_seconds

    #step.file_overwrite_policy
    #step.file_compare_type

    self.exclude_file_filter = step.exclude_file_filter
    self.include_file_filter = step.include_file_filter

    self.prefer_dest_initiated = step.prefer_dest_initiated

    self.datagram_size = step.datagram_size
    self.read_block_size = step.read_block_size
    self.write_block_size = step.write_block_size

    # clear ear, would be too confusing if password propagated from first step to template as a whole
    self.use_ear = false
    #self.ear_encrypted_password = ''
    self.ear_encrypted_password_confirm = ''

    # clear emails, would be too confusing if these got copied from first step
    self.emails=[]
  end

  def update_main_data(owner, job_template_data)
    self.description                  = job_template_data['description'].to_s.strip
    self.triggered                    = ['incoming_aspera', 'incoming_any'].include? job_template_data['source_type']

    self.user = owner
    self.source_type                        = job_template_data['source_type']
    self.encryption                         = job_template_data['encryption']
    self.target_rate                        = job_template_data['target_rate']
    self.minimum_rate                       = job_template_data['minimum_rate'].blank? ? 0 : job_template_data['minimum_rate']
    self.shared                             = job_template_data['shared']
    self.allow_initiator_changes            = job_template_data['allow_initiator_changes']

    self.console_tags                       = job_template_data['console_tags']

    self.delete_before_transfer             = job_template_data['delete_before_transfer']

    self.preserve_file_timestamps           = job_template_data['preserve_file_timestamps']
    self.preserve_file_create_times         = job_template_data['preserve_file_create_times']
    self.preserve_file_modify_times         = job_template_data['preserve_file_modify_times']
    self.preserve_file_access_times         = job_template_data['preserve_file_access_times']
    self.preserve_file_source_access_times  = job_template_data['preserve_file_source_access_times']
    self.preserve_file_uids                 = job_template_data['preserve_file_uids']
    self.preserve_file_gids                 = job_template_data['preserve_file_gids']

    self.symbolic_link_handling             = job_template_data['symbolic_link_handling']
    self.priority_and_policy                = job_template_data['priority_and_policy']

    self.use_ascp4                          = job_template_data['use_ascp4'] || false

    self.protocol                           = (Console.alternate_protocols_enabled ? job_template_data['protocol'] : 'FASP') || 'FASP'
    self.udp_port                           = job_template_data['udp_port']

    self.use_fasp_proxy                     = job_template_data['use_fasp_proxy']
    self.proxy_host                         = job_template_data['proxy_host']
    self.proxy_port                         = job_template_data['proxy_port']
    self.proxy_secure                       = job_template_data['proxy_secure']
    self.proxy_user                         = job_template_data['proxy_user']
    self.proxy_unencrypted_password         = job_template_data['proxy_unencrypted_password']

    self.file_exclude_by_timestamp          = job_template_data['file_exclude_by_timestamp']
    self.file_exclude_newer_than            = job_template_data['file_exclude_newer_than']

    self.max_retry_attempts                 = job_template_data['max_retry_attempts']
    self.retry_interval_seconds             = job_template_data['retry_interval_seconds']
    self.file_overwrite_policy              = job_template_data['file_overwrite_policy']
    self.file_compare_type                  = job_template_data['file_compare_type']
    self.exclude_file_filter                = job_template_data['exclude_file_filter']
    self.include_file_filter                = job_template_data['include_file_filter']

    self.prefer_dest_initiated              = job_template_data['prefer_dest_initiated']

    self.datagram_size                      = job_template_data['datagram_size']
    self.read_block_size                    = job_template_data['read_block_size']
    self.write_block_size                   = job_template_data['write_block_size']
    self.use_ear                            = job_template_data['use_ear']
    self.ear_encrypted_password             = job_template_data['ear_encrypted_password']
    self.ear_encrypted_password_confirm     = job_template_data['ear_encrypted_password_confirm']
    self.ear_unencrypted_password           = job_template_data['ear_encrypted_password']
    self.resume_partially_transferred_files = (job_template_data['resume_partially_transferred_files']==true)

    self.recurring                    = job_template_data['recurring']

    if job_template_data['recurring']
      job_template_data['schedule']['time_zone'] ||= owner.time_zone
      create_or_replace_schedule(job_template_data['schedule'])
    else
      self.schedule.destroy if self.schedule
    end

    job_template_data['emails'].each do |email_hash|
      if email_hash['start_template_id'].to_i > 0 || email_hash['success_template_id'].to_i > 0 || email_hash['error_template_id'].to_i > 0
        new_email = self.emails.build
        new_email.email = email_hash['email']
        new_email.send_email_on_start = email_hash['start_template_id'].to_i > 0
        new_email.send_email_on_success = email_hash['success_template_id'].to_i > 0
        new_email.send_email_on_error = email_hash['error_template_id'].to_i > 0
        new_email.start_template_id = email_hash['start_template_id'].to_i
        new_email.success_template_id = email_hash['success_template_id'].to_i
        new_email.error_template_id = email_hash['error_template_id'].to_i
        new_email.locked = email_hash['locked']
      end
    end

  end

  def create_or_replace_schedule(attrs)
    if self.schedule
      self.schedule.update_attributes(self.schedule.default_attributes.stringify_keys.merge(attrs))
    else
      self.build_schedule(attrs)
    end
    self.next_runtime = calculated_next_runtime
    self.schedule
  end


  def update_step_data(owner, job_template_data)
    job_template_data['job_template_steps'].each do |step_hash|
      step = self.job_template_steps.build
      # source login

      source_login_selected = owner.source_logins_for_create_smart.find {|x| x.id == step_hash['source_saved_login_id'].to_i}
      if source_login_selected
        if source_login_selected.connectable?
          step.source_login = source_login_selected
        else
          # need to create personal saved login, or edit existing if already exists
          unless new_source_login = owner.logins.first(:conditions => ["endpoint_type = ? AND login=? AND address=?", step_hash['source_endpoint_type'], step_hash['source_login'],step_hash['source_address']])
            new_source_login = Login.new
          end

          new_source_login.endpoint_type = step_hash['source_endpoint_type']
          new_source_login.address = step_hash['source_address']
          new_source_login.port = step_hash['source_port']
          new_source_login.login = step_hash['source_login']
          if step_hash['source_use_ssh_key']
            new_source_login.ssh_key_id = step_hash['source_ssh_key_id']
          else
            new_source_login.password = step_hash['source_password']
          end

          new_source_login.use_ssl = step_hash['source_use_ssl']
          new_source_login.validate_ssl_cert = step_hash['source_validate_ssl_cert']

          new_source_login.owner_type = 'User'
          new_source_login.owner_id = owner.id
          if new_source_login.valid?
            new_source_login.save!
          end
          step.source_login = new_source_login
        end
        step.source_node_id = (step.source_login.fasp_node && step.source_login.fasp_node.id)
      end

      # source directory
      step.source_directory = step_hash['source_directory']
      # source base
      step.source_base_enable = step_hash['source_base_enable']
      step.source_base = step_hash['source_base']

      # dest login
      dest_login_selected = owner.dest_logins_for_create_smart.find {|x| x.id == step_hash['dest_saved_login_id'].to_i}
      if dest_login_selected
        if dest_login_selected.connectable?
          step.dest_login = dest_login_selected
        else
          # need to create personal saved login, or edit existing if already exists
          unless new_dest_login = owner.logins.first(:conditions => ["endpoint_type = ? AND login=? AND address=?", step_hash['dest_endpoint_type'], step_hash['dest_login'],step_hash['dest_address']])
            new_dest_login = Login.new
          end
          new_dest_login.endpoint_type = step_hash['dest_endpoint_type']
          new_dest_login.address = step_hash['dest_address']
          new_dest_login.port = step_hash['dest_port']
          new_dest_login.login = step_hash['dest_login']
          if step_hash['dest_use_ssh_key']
            new_dest_login.ssh_key_id = step_hash['dest_ssh_key_id']
          else
            new_dest_login.password = step_hash['dest_password']
          end

          new_dest_login.use_ssl = step_hash['dest_use_ssl']
          new_dest_login.validate_ssl_cert = step_hash['dest_validate_ssl_cert']

          new_dest_login.owner_type = 'User'
          new_dest_login.owner_id = owner.id
          if new_dest_login.valid?
            new_dest_login.save!
          end
          step.dest_login = new_dest_login
        end
        step.dest_node_id = (step.dest_login.fasp_node && step.dest_login.fasp_node.id)
      end


      # dest directory
      step.dest_directory = step_hash['dest_directory']

      # initiator allowed to change dest path
      step.allow_change_dest_dir = step_hash['allow_change_dest_dir']==true

      # initiator allowed to remove dest
      step.allow_delete_dest = step_hash['allow_delete_dest']==true

      # override template defaults
      step.override_template_defaults = step_hash['override_template_defaults']

      step.use_ascp4 = step_hash['use_ascp4'] || false

      step.archive_after_transfer = step_hash['archive_after_transfer'] == true
      step.archive_directory = step_hash['archive_directory']
      step.delete_source_after_transfer = (step_hash['delete_source_after_transfer']==true)
      step.delete_source_dirs_after_transfer = (step_hash['delete_source_dirs_after_transfer']==true)
      step.delete_empty_directories = (step_hash['delete_empty_directories']==true)
      # step.dest_node

      # step.source_node
      step.source_is_url = step_hash['source_is_url']
      step.dest_is_url = step_hash['dest_is_url']

      if step.override_template_defaults
        # get from job step hash
        step.target_rate = step_hash['target_rate']
        step.minimum_rate = step_hash['minimum_rate'].blank? ? 0 : step_hash['minimum_rate']
        step.resume_partially_transferred_files = (step_hash['resume_partially_transferred_files']==true)

        step.console_tags = step_hash['console_tags']

        step.delete_before_transfer = step_hash['delete_before_transfer']

        step.preserve_file_timestamps = step_hash['preserve_file_timestamps']
        step.preserve_file_create_times = step_hash['preserve_file_create_times']
        step.preserve_file_modify_times = step_hash['preserve_file_modify_times']
        step.preserve_file_access_times = step_hash['preserve_file_access_times']
        step.preserve_file_source_access_times = step_hash['preserve_file_source_access_times']
        step.preserve_file_uids = step_hash['preserve_file_uids']
        step.preserve_file_gids = step_hash['preserve_file_gids']

        step.symbolic_link_handling = step_hash['symbolic_link_handling']
        step.priority_and_policy = step_hash['priority_and_policy']

        step.use_ascp4 = step_hash['use_ascp4'] || false

        step.protocol = (Console.alternate_protocols_enabled ? step_hash['protocol'] : 'FASP') || 'FASP'
        step.udp_port = step_hash['udp_port']
        step.use_fasp_proxy = step_hash['use_fasp_proxy']
        step.proxy_host = step_hash['proxy_host']
        step.proxy_port = step_hash['proxy_port']
        step.proxy_secure = step_hash['proxy_secure']
        step.proxy_user = step_hash['proxy_user']
        step.proxy_unencrypted_password = step_hash['proxy_unencrypted_password']

        step.file_exclude_by_timestamp = step_hash['file_exclude_by_timestamp']
        step.file_exclude_newer_than = step_hash['file_exclude_newer_than']

        step.max_retry_attempts = step_hash['max_retry_attempts']
        step.retry_interval_seconds = step_hash['retry_interval_seconds']
        step.file_overwrite_policy = step_hash['file_overwrite_policy']
        step.file_compare_type = step_hash['file_compare_type']
        step.exclude_file_filter = step_hash['exclude_file_filter']
        step.include_file_filter = step_hash['include_file_filter']

        step.prefer_dest_initiated = step_hash['prefer_dest_initiated']

        step.datagram_size = step_hash['datagram_size']
        step.read_block_size = step_hash['read_block_size']
        step.write_block_size = step_hash['write_block_size']
        step.use_ear = step_hash['use_ear']
        step.ear_encrypted_password = step_hash['ear_encrypted_password']
        step.ear_encrypted_password_confirm = step_hash['ear_encrypted_password_confirm']
        step.ear_unencrypted_password = step_hash['ear_encrypted_password']
        step.encryption = step_hash['encryption']
        # emails
        step_hash['emails'].each do |email_hash|
          if email_hash['start_template_id'].to_i > 0 || email_hash['success_template_id'].to_i > 0 || email_hash['error_template_id'].to_i > 0
            new_email = step.emails.build
            new_email.email = email_hash['email']
            new_email.send_email_on_start = email_hash['start_template_id'].to_i > 0
            new_email.send_email_on_success = email_hash['success_template_id'].to_i > 0
            new_email.send_email_on_error = email_hash['error_template_id'].to_i > 0
            new_email.start_template_id = email_hash['start_template_id'].to_i
            new_email.success_template_id = email_hash['success_template_id'].to_i
            new_email.error_template_id = email_hash['error_template_id'].to_i
            new_email.locked = email_hash['locked']
          end
        end
      else
        # copy from job template
        step.target_rate = self.target_rate
        step.minimum_rate = self.minimum_rate || 0
        step.resume_partially_transferred_files = self.resume_partially_transferred_files

        step.console_tags = self.console_tags

        step.delete_before_transfer = self.delete_before_transfer

        step.preserve_file_timestamps = self.preserve_file_timestamps
        step.preserve_file_create_times = self.preserve_file_create_times
        step.preserve_file_modify_times = self.preserve_file_modify_times
        step.preserve_file_access_times = self.preserve_file_access_times
        step.preserve_file_source_access_times = self.preserve_file_source_access_times
        step.preserve_file_uids = self.preserve_file_uids
        step.preserve_file_gids = self.preserve_file_gids

        step.symbolic_link_handling = self.symbolic_link_handling
        step.priority_and_policy = self.priority_and_policy

        step.use_ascp4 = self.use_ascp4

        step.protocol = (Console.alternate_protocols_enabled ? self.protocol : 'FASP') || 'FASP'
        step.udp_port = self.udp_port
        step.use_fasp_proxy = self.use_fasp_proxy
        step.proxy_host = self.proxy_host
        step.proxy_port = self.proxy_port
        step.proxy_secure = self.proxy_secure
        step.proxy_user = self.proxy_user
        step.proxy_unencrypted_password = self.proxy_unencrypted_password

        step.file_exclude_by_timestamp = self.file_exclude_by_timestamp
        step.file_exclude_newer_than = self.file_exclude_newer_than

        step.max_retry_attempts = self.max_retry_attempts
        step.retry_interval_seconds = self.retry_interval_seconds
        step.file_overwrite_policy = self.file_overwrite_policy
        step.file_compare_type = self.file_compare_type
        step.exclude_file_filter = self.exclude_file_filter
        step.include_file_filter = self.include_file_filter

        step.prefer_dest_initiated = self.prefer_dest_initiated

        step.datagram_size = self.datagram_size
        step.read_block_size = self.read_block_size
        step.write_block_size = self.write_block_size
        step.use_ear = self.use_ear
        step.ear_encrypted_password = self.ear_encrypted_password
        step.ear_encrypted_password_confirm = self.ear_encrypted_password_confirm
        step.ear_unencrypted_password = self.ear_unencrypted_password
        step.encryption = self.encryption
        #emails
        job_template_data['emails'].each do |email_hash|
          if email_hash['start_template_id'].to_i > 0 || email_hash['success_template_id'].to_i > 0 || email_hash['error_template_id'].to_i > 0
            new_email = step.emails.build
            new_email.email = email_hash['email']
            new_email.send_email_on_start = email_hash['start_template_id'].to_i > 0
            new_email.send_email_on_success = email_hash['success_template_id'].to_i > 0
            new_email.send_email_on_error = email_hash['error_template_id'].to_i > 0
            new_email.start_template_id = email_hash['start_template_id'].to_i
            new_email.success_template_id = email_hash['success_template_id'].to_i
            new_email.error_template_id = email_hash['error_template_id'].to_i
            new_email.locked = email_hash['locked']
          end
        end
      end

      step.source_initiated = Login::choose_source_initiated(step.source_login, step.dest_login, step.prefer_dest_initiated)

      # Non-initiator must not be node API
      source_suitable = step.source_login.present? && step.source_login.fasp_node.present?
      dest_suitable = step.dest_login.present? && step.dest_login.fasp_node.present?
    end
  end

  def repeat_minute
    "%02d" % self['repeat_minute']
  end

  def cancel_scheduled_transfers
    err_desc = "Cancelled due to delete of smart transfer template"
    steps = jobs.map(&:steps).flatten.select { |step| step.status_scheduled? }
    steps.each do |step|
      step.request_cancel_never_started(err_desc)
    end
  end

  # private

  def stripped_value(string)
    string.nil? ? "" : string.strip
  end

  def select_sources?
    source_type == 'user_selected'
  end
end
